/* eslint-env node */
'use strict';
// Reference: http://www.accessreading.org/api.html


let HTTPRequest = null;
let networkCache = null;

// このモジュールを読込んだら、必ず以下のメソッドを使ってXMLHttpRequestを登録する
exports.setXMLHttpRequest = function (xmlHttpRequest) {
	HTTPRequest = xmlHttpRequest;
	return exports;
};

// このモジュールを読込んだら、必ず以下のメソッドを使ってキャッシュ用のオブジェクトを登録する
exports.setNetworkCache = function (nc) {
	networkCache = nc;
	return exports;
};


function isObject (arg) {
	return arg && ('object' === typeof arg || 'function' === typeof arg);
}


const URL4AR = exports.URL = 'https://books.accessreading.org/';
const formatTypes = exports.formatTypes = ['docx', 'epub'];


// デフォルトのタイムアウトをミリ秒で指定
const TIMEOUT = 16000;

// HTTPの定数を定義
const HTTP_OK = 200;
const HTTP_UNAUTHORIZED = 401;
const HTTP_FORBIDDEN = 403;
const HTTP_NOT_FOUND = 404;

// エラーの説明文を定義
const errorDescription = {
	error: '通信が失敗しました',
	abort: '通信が中断されました',
	timeout: '通信が時間切れにより止まりました'
};


/*
	検索APIに使うキーと値からなるオブジェクトから適宜パーセントエンコードされた文字列を出力
		例: object2encodedParameter({Author: 'hoge', title: 'piyo あ'}, 2) -> '?Page=2&title=piyo+%E3%81%82&Author=hoge'
*/
function object2encodedParameter (ao, pn) {
	const argObject = isObject(ao) ? ao : {};
	const pageNumber = 'number' === typeof pn ? pn : 1;
	return ['title', 'Author', 'CategoryId', 'Publisher', 'GradeId', 'BookId', 'downloadable'].reduce(function (previousValue, currentName) {
		let encodedParameter = previousValue;
		if (currentName in argObject) {
			const strPropertyValue = String(argObject[currentName]);
			if (strPropertyValue.length) {
				encodedParameter += '&' + currentName + '=' + encodeURIComponent(strPropertyValue).replace('%20', '+');
			}
		}
		return encodedParameter;
	}, '?Page=' + pageNumber);
}


exports.User = class {
	constructor (username, password) {
		this.username = username;
		this.password = password;
		if (!isObject(networkCache[username])) {
			networkCache[username] = {};
		}
		this.cache = networkCache[username];
	}


	sendRequest (pathText, onLoadEnd, noAuth, mn, isArrayBuffer, timeout) {
		const self = this;
		return new Promise(function (resolve, reject) {
			const req = new HTTPRequest();
			const url = URL4AR + pathText;
			const methodName = mn ? mn : 'GET';

			if (noAuth) {
				req.open(methodName, url);
			} else {
				req.open(methodName, url, true, self.username, self.password);
				req.setRequestHeader('Authorization', 'Basic ' + (new Buffer(self.username + ':' + self.password)).toString('base64')); // この行は省略できるはずだが、そうすると複数のアカウントで同時アクセスしたときにおかしくなることがある
			}

			req.timeout = timeout ? timeout : TIMEOUT;
			req.responseType = isArrayBuffer ? 'arraybuffer' : 'json';

			req.addEventListener('load', function () {
				if ([HTTP_OK, HTTP_UNAUTHORIZED, HTTP_FORBIDDEN, HTTP_NOT_FOUND].indexOf(req.status) + 1) {
					// HTTPステータスコードが200でなくても、レスポンスボディがJSONで返る可能性がある場合は取敢えずonFulfilled扱いに
					resolve(req);
				} else {
					reject(new Error(`"${url}"への通信が次の理由により失敗しました。${req.statusText}`));
				}
			});
			Object.keys(errorDescription).forEach(function (eventType) {
				req.addEventListener(eventType, function () {
					reject(new Error(`"${url}"への${errorDescription[eventType]}。${req.statusText}`));
				});
			});
			if ('function' === typeof onLoadEnd) {
				req.addEventListener('loadend', onLoadEnd);
			}

			req.send();
		});
	}


	// 書籍情報検索結果取得
	searchBooks (parameterObject, onLoadEnd) {
		return this.sendRequest('api/search/books' + object2encodedParameter(parameterObject), onLoadEnd).then(function (req) {
			if (HTTP_OK === req.status) {
				return req.response;
			} else if (HTTP_FORBIDDEN === req.status) {
				return Promise.reject(new Error(`検索クエリ${JSON.stringify(parameterObject)}の送信が次の理由により失敗しました。${req.statusText}`));
			}
			// ここに到達する場合(req.statusがHTTP_UNAUTHORIZEDかHTTP_NOT_FOUND)、レスポンスボディがJSONで返ることが確実なので、それをエラーオブジェクトに反映させる
			return Promise.reject(new Error(`検索クエリ${JSON.stringify(parameterObject)}の送信が次の理由により失敗しました。${req.response.Message}`));
		});
	}

	// ダウンロード可能な書籍一覧を返す
	searchBooksDownloadable (onLoadEnd) {
		const userCache = this.cache;
		if (userCache.books) {
			if (onLoadEnd) {
				if ('function' !== typeof onLoadEnd) {
					throw new TypeError();
				}
				onLoadEnd();
			}
			return Promise.resolve(userCache.books);
		}
		return this.searchBooks({downloadable: 1}, onLoadEnd).then(function (responseJSON) {
			userCache.books = {};
			responseJSON.forEach(function (book) {
				if (!book.HasBookFiles) {	return;	}
				userCache.books[book.BookId] = book;
			});
			return userCache.books;
		});
	}


	// 書籍のサムネイルの有無を返す
	checkCoverimage (bookID, onLoadEnd) {
		const bookCache = this.cache.books[bookID];
		if ('hasCoverimage' in bookCache) {
			if (onLoadEnd) {
				if ('function' !== typeof onLoadEnd) {
					throw new TypeError();
				}
				onLoadEnd();
			}
			return Promise.resolve(bookCache.hasCoverimage);
		}
		return this.sendRequest('Image/Index/' + bookID, onLoadEnd, true, 'HEAD').then(function (req) {
			bookCache.hasCoverimage = HTTP_OK === req.status;
			return bookCache.hasCoverimage;
		}).catch(function (e) {
			return Promise.reject(new Error(`BookID: ${bookID} である書籍の表紙画像の有無の確認に失敗しました。${e.message}`));
		});
	}


	// 書籍ファイル一覧取得
	getBookfiles (bookID, onLoadEnd) {
		const bookCache = this.cache.books[bookID];
		let bookfileCache = this.cache.bookfiles;
		if (!isObject(bookfileCache)) {
			bookfileCache = this.cache.bookfiles = {};
		}
		if (bookCache.bookfiles) {
			if (onLoadEnd) {
				if ('function' !== typeof onLoadEnd) {
					throw new TypeError();
				}
				onLoadEnd();
			}
			return Promise.resolve(bookCache.bookfiles);
		}
		return this.sendRequest('api/bookfiles/' + bookID, onLoadEnd).then(function (req) {
			if (HTTP_OK === req.status) {
				const bookfiles = bookCache.bookfiles = {};
				formatTypes.forEach(function (type) {
					bookfiles[type] = {
						length: 0,
						hasRuby: {length: 0, hasPicture: {length: 0}, hasNoPicture: {length: 0}},
						hasNoRuby: {length: 0, hasPicture: {length: 0}, hasNoPicture: {length: 0}}
					};
				});
				req.response.forEach(function (bf) {
					let destination = bookfiles[bf.FileFormat.toLowerCase()];
					destination.length += 1;
					destination = destination[bf.RubiExists ? 'hasRuby' : 'hasNoRuby'];
					destination.length += 1;
					destination = destination[bf.ImageExists ? 'hasPicture' : 'hasNoPicture'];
					destination.length += 1;
					bookfileCache[bf.BookFileId] = destination[bf.BookFileId] = bf;
				});
				return bookfiles;
			}
			return Promise.reject(new Error(`BookID: ${bookID} に紐づけられたファイル情報の取得が次の理由により失敗しました。${req.response.Message}`));
		});
	}


	// 書籍ファイルリクエスト
	requestBookfile (bookfileID, onLoadEnd) {
		const cache = this.cache.bookfiles[bookfileID];
		if (cache.hasRequested) {
			if (onLoadEnd) {
				if ('function' !== typeof onLoadEnd) {
					throw new TypeError();
				}
				onLoadEnd();
			}
			return Promise.resolve(cache.hasRequested);
		}
		return this.sendRequest('api/book/' + bookfileID, onLoadEnd, false, 'PUT').then(function (req) {
			if (HTTP_OK === req.status || HTTP_NOT_FOUND === req.status) { // HTTP_NOT_FOUNDの場合リクエスト済の場合があり得るので成功扱い
				cache.hasRequested = true;
				return cache.hasRequested;
			}
			return Promise.reject(new Error(`BookFileID: ${bookfileID} へのダウンロード要求が次の理由により失敗しました。${req.response.Message}`));
		});
	}

	// 書籍ファイルダウンロード
	downloadBookfile (bookfileID, onLoadEnd, ttl) {
		if (!ttl) { // 何らかの理由で...->requestBookfile->downloadBookfile->requestBookfile->...が無限ループに陥ることを防ぐ
			return Promise.reject(new Error(`BookFileID: ${bookfileID} へのダウンロード試行回数が限界を越えました。`));
		}
		const self = this;
		const cache = self.cache.bookfiles[bookfileID];
		/* if (cache.hasDownloaded) {
			if (onLoadEnd) {
				if ('function' !== typeof onLoadEnd) {
					throw new TypeError();
				}
				onLoadEnd();
			}
			return Promise.resolve(cache.hasDownloaded);
		} */
		if (!cache.hasRequested) {
			return self.requestBookfile(bookfileID).then(function () {
				return self.downloadBookfile(bookfileID, onLoadEnd, ttl - 1);
			});
		}
		return self.sendRequest('api/book/' + bookfileID, onLoadEnd, false, 'GET', true, 1024000).then(function (req) {
			if (HTTP_OK === req.status) {
				cache.hasDownloaded = true;
				return new Buffer(new Uint8Array(req.response));
			}
			return Promise.reject(new Error(`BookFileID: ${bookfileID} のダウンロードが次の理由により失敗しました。${req.response.Message}`));
		});
	}
};
