(function () {
	'use strict';
	/* eslint-env browser */
	/* global require Buffer */
	/* eslint global-require: 0 */

	// 各種ユーティリティ
	function isObject (arg) {
		return arg && ('object' === typeof arg || 'function' === typeof arg);
	}
	function getClassName (arg) {
		if (!isObject(arg)) {
			throw new TypeError();
		}
		return Object.prototype.toString.call(arg).replace(/^\[object (\w+)\]$/, '$1');
	}
	function object2map (arg) {
		if (!isObject(arg)) {
			throw new TypeError();
		}
		return Object.keys(arg).reduce(function (map, propertyName) {
			return map.set(propertyName, arg[propertyName]);
		}, new Map());
	}
	function likesNonEmptyString (arg) {
		return arg || 0 === arg;
	}

	// NW.jsのAPIに関する変数の定義
	const gui = require('nw.gui');
	const shell = gui.Shell;
	const manifest = gui.App.manifest;

	const filesystem = require('fs');
	const path = require('path');

	// 書籍ファイルのダウンロード先ディレクトリがなければ作る
	const path2bookshelves = path.join(gui.App.dataPath, 'bookshelves');
	let bookshelvesExists = false;
	try {
		bookshelvesExists = filesystem.statSync(path2bookshelves).isDirectory();
	} catch (e) {
		filesystem.mkdirSync(path2bookshelves);
		bookshelvesExists = true;
	}
	if (!bookshelvesExists) {
		throw new Error();
	}

	// # DOM操作関数の定義
	function getElementById (id) {
		return document.getElementById(id);
	}
	// createElementNS(名前空間, 要素名, {attributes: {属性名: 属性値}, parent: 親要素, child: 子要素, children: [子要素, 子要素, ...]})
	function createElementNS (namespace, tagName, parameters) {
		const newElement = document.createElementNS(namespace, tagName);
		if (!parameters) {
			return newElement;
		}

		if (!isObject(parameters)) {
			throw new TypeError();
		}
		const attributes = parameters.attributes;
		if (attributes) {
			if (!isObject(attributes)) {
				throw new TypeError();
			}
			Object.keys(attributes).forEach(function (attributeName) {
				newElement[attributeName] = attributes[attributeName];
			});
		}
		const parentElement = parameters.parent;
		if (parentElement) {
			if (!isObject(parentElement) || !getClassName(parentElement).match(/^\w*Element$/)) {
				throw new TypeError();
			}
			parentElement.append(newElement);
		}
		const child2append = parameters.child;
		let children2append = parameters.children;
		if (children2append) {
			if (!Array.isArray(children2append)) {
				throw new TypeError();
			}
		} else {
			children2append = [];
		}
		if (child2append) {
			children2append.unshift(child2append);
		}
		children2append.forEach(function (child) {
			if (!(isObject(child) && getClassName(child).match(/^(Text|\w*Element)$/) || 'number' === typeof child || 'string' === typeof child)) {
				throw new TypeError();
			}
			newElement.append(child);
		});
		return newElement;
	}

	// ## XHTML関連
	const createXHTMLElement = createElementNS.bind(null, document.documentElement.namespaceURI);
	const createForm = createXHTMLElement.bind(null, 'form');
	const createInput = createXHTMLElement.bind(null, 'input');
	const createOutput = createXHTMLElement.bind(null, 'output');
	const createParagraph = createXHTMLElement.bind(null, 'p');
	const createButton = createXHTMLElement.bind(null, 'button');
	const createLabel = createXHTMLElement.bind(null, 'label');
	const createHeading = createXHTMLElement.bind(null, 'h1');
	function createImage (parameters) {
		const image = createXHTMLElement('img', parameters);
		if (!likesNonEmptyString(image.title)) {	image.title = image.alt;	}
		return image;
	}
	function createAnchor (parameters, outerURL) {
		const anchor = createXHTMLElement('a', parameters);
		if (!likesNonEmptyString(anchor.href)) {
			anchor.href = '#';
		}
		if (outerURL) {
			if (!likesNonEmptyString(anchor.title)) {
				anchor.title = outerURL;
			}
			anchor.addEventListener('click', function () {
				if (confirm(`外部サイト "${outerURL}" を表示します`)) {
					shell.openExternal(outerURL);
				}
			});
		}
		return anchor;
	}
	function createSection (parameters, headingText) {
		const section = createXHTMLElement('section', parameters);
		if (likesNonEmptyString(headingText)) {
			createHeading({parent: section, child: headingText});
		}
		return section;
	}
	function map2definitionList (arg, parameters) {
		if (!isObject(arg)) {
			throw new TypeError();
		}
		const map = 'Map' === getClassName(arg) ? arg : object2map(arg);
		const dl = createXHTMLElement('dl', parameters);
		for (const key of map.keys()) {
			createXHTMLElement('dt', {parent: dl, child: key});
			createXHTMLElement('dd', {parent: dl, child: map.get(key)});
		}
		return dl;
	}

	// ## SVG関連
	const SVG_NAMESPACE = 'http://www.w3.org/2000/svg';
	const SVG_VERSION = 1.1;
	const createSVGElement = createElementNS.bind(null, SVG_NAMESPACE);


	function createMenuIcon (title, drawPath) {
		const svgElement = createSVGElement('svg', {children: [
			createSVGElement('title', {child: title}),
			createSVGElement('g')
		]});
		svgElement.setAttribute('version', SVG_VERSION);
		svgElement.setAttribute('viewBox', '0 0 17 17');
		createSVGElement('path', {parent: svgElement}).setAttribute('d', drawPath);
		return svgElement;
	}


	let lastSelected = localStorage.getItem('lastSelected');
	if (!lastSelected) {
		lastSelected = 'newUser';
	}

	const crypto = require('crypto');
	const magicBuffer = new Buffer('f6389a853f2c538ad37cf23dac13c0da75dfd5c1c4098da21329b1639fcc8bd2', 'hex');
	let applicationKey = localStorage.getItem('applicationKey');
	if (applicationKey) {
		const decipher = crypto.createDecipher('aes256', magicBuffer);
		applicationKey = Buffer.concat([decipher.update(applicationKey, 'hex'), decipher.final()]);
	} else {
		applicationKey = '';
		for (let index = 0; index < 64; index += 1) {
			applicationKey += Math.floor(Math.random() * 16).toString(16);
		}
		applicationKey = new Buffer(applicationKey, 'hex');
		const cipher = crypto.createCipher('aes256', magicBuffer);
		localStorage.setItem(
			'applicationKey',
			cipher.update(applicationKey, null, 'hex') + cipher.final('hex')
		);
	}
	function loadPasswords () {
		const cipheredPasswords = JSON.parse(localStorage.getItem('passwords'));
		if (!isObject(cipheredPasswords)) {
			return {};
		}
		return Object.keys(cipheredPasswords).reduce(function (passwords, username) {
			const decipher = crypto.createDecipher(
				'aes256',
				crypto.createHash('sha256').update(username, 'utf8').update(applicationKey).digest()
			);
			passwords[username] = decipher.update(cipheredPasswords[username], 'hex', 'utf8') + decipher.final('utf8');
			return passwords;
		}, {});
	}
	const passwords = loadPasswords();
	function savePasswords () {
		localStorage.setItem('passwords', JSON.stringify(Object.keys(passwords).reduce(function (cipheredPasswords, username) {
			const cipher = crypto.createCipher(
				'aes256',
				crypto.createHash('sha256').update(username, 'utf8').update(applicationKey).digest()
			);
			cipheredPasswords[username] = cipher.update(passwords[username], 'utf8', 'hex') + cipher.final('hex');
			return cipheredPasswords;
		}, {})));
	}

	let nicknames = JSON.parse(localStorage.getItem('nicknames'));
	if (!isObject(nicknames)) {
		nicknames = {};
	}

	let fileformats = JSON.parse(localStorage.getItem('fileformats'));
	if (!isObject(fileformats)) {
		fileformats = {};
	}

	let networkCache = JSON.parse(localStorage.getItem('networkCache'));
	if (!isObject(networkCache)) {
		networkCache = {};
	}

	// ウィンドウが閉じられる前に、各種変数をlocalStorageへ保存する。但し、全情報リセットの場合はその限りではない。
	let isReset = false;
	const nwWindow = gui.Window.get();
	let onQuit = null;
	function onClose () {
		if (isReset) {	return;	}
		const localStorage = nwWindow.window.localStorage;
		localStorage.setItem('lastSelected', lastSelected);
		localStorage.setItem('nicknames', JSON.stringify(nicknames));
		localStorage.setItem('fileformats', JSON.stringify(fileformats));
		localStorage.setItem('networkCache', JSON.stringify(networkCache));
		savePasswords();
		nwWindow.removeListener('close', onQuit);
	}
	onQuit = function () {
		try {	onClose();	} catch (e) {	alert(e.message);	}
		this.close(true);
	};
	window.addEventListener('beforeunload', onClose);
	nwWindow.on('close', onQuit);


	const AccessReading = require('./accessreading.js').setXMLHttpRequest(XMLHttpRequest).setNetworkCache(networkCache);
	const ARUser = AccessReading.User;
	// 読込中を表すSVGを埋め込む関数
	function createLoadingIcon (parentElement) {
		const svgElement = createSVGElement('svg', {child: createSVGElement('title', {child: '通信中'})});
		svgElement.setAttribute('version', SVG_VERSION);
		svgElement.setAttribute('viewBox', '0 0 155 140');
		function createRectangle (x, y, height, begin) {
			const rect = createSVGElement('rect', {parent: svgElement});
			rect.setAttribute('x', x);
			rect.setAttribute('y', y);
			rect.setAttribute('width', '15');
			rect.setAttribute('height', height);
			rect.setAttribute('rx', '6');
			function createAnimation (attributeName, values) {
				const animate = createSVGElement('animate', {parent: rect});
				animate.setAttribute('attributeName', attributeName);
				animate.setAttribute('begin', begin);
				animate.setAttribute('dur', '1s');
				animate.setAttribute('values', values);
				animate.setAttribute('repeatCount', 'indefinite');
			}
			createAnimation('height', '120;110;100;90;80;70;60;50;40;140;120');
			createAnimation('y', '10;15;20;25;30;35;40;45;50;0;10');
		}
		createRectangle('20', '10', '120', '0.5s');
		createRectangle('50', '10', '120', '0.25s');
		createRectangle('80', '0', '140', '0s');
		createRectangle('110', '10', '120', '0.25s');
		createRectangle('140', '10', '120', '0.5s');
		parentElement.append(svgElement);
		return svgElement;
	}


	const paths = require('./paths2draw.json');
	const navList = document.query('header nav ul');
	const mainElement = document.query('main');
	const formatTypes = AccessReading.formatTypes;
	const tabs = {};
	class TabbedPanel {
		constructor (str, title) {
			const self = this;

			self.listItem = createXHTMLElement('li', {parent: navList});

			self.anchor = createAnchor();
			self.anchor.addEventListener('click', function () {
				if (lastSelected in tabs) {
					tabs[lastSelected].unselect();
				}
				self.select();
			});

			if (title) {
				self.createPanelWithIcon(str, title);
			} else {
				self.createBookshelf(str);
			}

			tabs[self.sectionID] = self;
		}


		createPanelWithIcon (sectionID, title) {
			this.sectionID = sectionID;
			this.listItem.title = title;
			this.listItemContent = createMenuIcon(title, paths[sectionID]);
			this.contentSection = createSection({parent: mainElement}, title);
		}


		createBookshelf (username) {
			const self = this;
			self.sectionID = 'bookshelf_' + username;
			self.username = self.listItem.title = username;
			const alias = nicknames[username];
			self.busyCounter = 0;
			self.listItemContent = createXHTMLElement('span', {child: alias});
			const bookshelf = self.contentSection = createSection({parent: mainElement});
			const anchorInHeading = createAnchor({attributes: {title: `ユーザー "${username}" の表示名を変える`}, child: alias});
			anchorInHeading.addEventListener('click', function () {
				let newAlias = prompt(`ユーザー "${username}" の新しい表示名を入力`, alias);
				if (null === newAlias) {	return;	}
				newAlias = newAlias.trim();
				if (alias === newAlias) {	return;	}
				if (!newAlias.length) {
					alert('無効な名前です');
					return;
				}
				nicknames[username] = newAlias;
				location.reload();
			});
			createHeading({parent: bookshelf, children: [anchorInHeading, 'の本棚']});
			const formatChooser = createForm({
				parent: bookshelf,
				child: createParagraph({children: formatTypes.map(function (format) {
					return createLabel({children: [
						format,
						createInput({attributes: {name: 'fileformat', type: 'radio', value: format}})
					]});
				})
			})});
			formatChooser.fileformat.value = fileformats[username];
			formatChooser.addEventListener('change', function () {
				fileformats[username] = formatChooser.fileformat.value;
				formatTypes.forEach(function (format) {
					if (format === fileformats[username]) {
						getElementById(`bookshelf_${username}_${format}`).classList.remove('unselected');
					} else {
						getElementById(`bookshelf_${username}_${format}`).classList.add('unselected');
					}
				});
			});

			self.arUser = new ARUser(username, passwords[username]);
			self.increaseCounter();
			self.arUser.searchBooksDownloadable(function () {
				self.decreaseCounter();
			}).then(function (downloadableBooks) {
				formatTypes.forEach(function (format) {
					self.createBookSection(downloadableBooks, format, createSection({
						parent: bookshelf,
						attributes: {id: `bookshelf_${username}_${format}`}
					}));
				});
				formatChooser.dispatchEvent(new Event('change'));
				self.createRemoveButton();
			}).catch(function (e) {
				createParagraph({parent: bookshelf, child: e.message});
				self.createRemoveButton();
			});
		}

		createRemoveButton () {
			const username = this.username;
			const nickname = nicknames[username];
			createForm({attributes: {method: 'post'}, parent: this.contentSection, child: createParagraph({
				child: createButton({child: `"${nickname}"の情報を全て消す`})
			})}).addEventListener('submit', function (event) {
				if (confirm(`"${username}" というユーザー${username === nickname ? '' : ` (表示名: "${nickname}") `}の情報を全て消します。本当によろしいですか?`)) {
					lastSelected = 'newUser';
					delete passwords[username];
					delete nicknames[username];
					delete fileformats[username];
					delete networkCache[username];
					return;
				}
				event.preventDefault();
			});
		}


		createBookSection (downloadableBooks, format, parentSection) {
			const self = this;
			const arUser = self.arUser;
			Object.keys(downloadableBooks).forEach(function (bookID) {
				self.increaseCounter();
				arUser.getBookfiles(bookID, function () {
					self.decreaseCounter();
				}).then(function (bookfiles) {
					const bookfilesSameFormat = bookfiles[format];
					if (!bookfilesSameFormat.length) {	return;	}
					self.increaseCounter();
					arUser.checkCoverimage(bookID, function () {
						self.decreaseCounter();
					}).then(function (coverimageExists) {
						const book = downloadableBooks[bookID];
						const bookSection = createSection({parent: parentSection}, book.Title);
						if (coverimageExists) {
							const paragraph4coverimage = createParagraph({
								parent: bookSection,
								child: createImage({attributes: {src: AccessReading.URL + 'Image/Index/' + bookID, alt: book.Title + 'の表紙画像', width: 128}})
							});
							paragraph4coverimage.classList.add('coverimage');
						}
						map2definitionList({'出版社': book.Publisher}, {parent: bookSection});
						function listFilesR (bfr, parent) {
							function listFilesP (files, p) {
								Object.keys(files).forEach(function (bookfileID) {
									if ('length' === bookfileID) {	return;	}
									const bookfileInfo = arUser.cache.bookfiles[bookfileID];
									const button2read = createButton({child: '読む'});
									const button2delete = createButton({child: '消す'});
									const button2download = createButton({child: 'ダウンロード'});
									const buttons = createParagraph({parent: p, child: bookfileInfo.Title + ' '});
									button2read.addEventListener('click', function () {
										shell.openItem(path.join(path2bookshelves, self.username, bookfileID + '.' + format));
									});
									button2delete.addEventListener('click', function () {
										button2read.remove();
										button2delete.remove();
										if (confirm(`${bookfileInfo.Title}のファイルを消します。本当によろしいですか?`)) {
											filesystem.unlink(path.join(path2bookshelves, self.username, bookfileID + '.' + format), function (err) {
												if (err) {	throw err;	}
												bookfileInfo.hasDownloaded = false;
												buttons.append(button2download);
											});
										} else {
											buttons.append(button2read, ' ', button2delete);
										}
									});
									button2download.addEventListener('click', function () {
										button2download.remove();
										self.increaseCounter();
										arUser.downloadBookfile(bookfileID, function () {
											self.decreaseCounter();
										}, 4).then(function (buf) {
											filesystem.writeFile(
												path.join(path2bookshelves, self.username, bookfileID + '.' + format),
												buf,
												function (err) {
													if (err) {	throw err;	}
													buttons.append(button2read, ' ', button2delete);
												}
											);
										}).catch(function () {
											buttons.append(button2download);
										});
									});
									filesystem.stat(path.join(path2bookshelves, self.username, bookfileID + '.' + format), function (err, stats) {
										if (err) {
											bookfileInfo.hasDownloaded = false;
										} else if (stats.isFile()) {
											bookfileInfo.hasDownloaded = true;
										} else {
											throw new Error();
										}
										if (bookfileInfo.hasDownloaded) {
											buttons.append(button2read, ' ', button2delete);
										} else {
											buttons.append(button2download);
										}
									});
								});
							}
							if (bfr.hasPicture.length) {
								listFilesP(bfr.hasPicture, createSection({parent: parent}, '挿絵あり'));
								if (bfr.hasNoPicture.length) {
									listFilesP(bfr.hasNoPicture, createSection({parent: parent}, '挿絵なし'));
								}
							} else {
								listFilesP(bfr.hasNoPicture, parent);
							}
						}
						if (bookfilesSameFormat.hasRuby.length) {
							listFilesR(bookfilesSameFormat.hasRuby, createSection({parent: bookSection}, '振仮名あり'));
							if (bookfilesSameFormat.hasNoRuby.length) {
								listFilesR(bookfilesSameFormat.hasNoRuby, createSection({parent: bookSection}, '振仮名なし'));
							}
						} else {
							listFilesR(bookfilesSameFormat.hasNoRuby, bookSection);
						}
					});
				});
			});
		}


		increaseCounter () {
			if (!this.busyCounter) {
				this.loadingIcon = createLoadingIcon(this.listItemContent);
			}
			this.busyCounter += 1;
		}

		decreaseCounter () {
			this.busyCounter -= 1;
			if (this.busyCounter <= 0) {
				this.busyCounter = 0;
				this.loadingIcon.remove();
			}
		}


		select () {
			this.listItem.classList.add('selected');
			this.contentSection.classList.add('selected');

			this.listItemContent.remove();
			this.anchor.remove();
			this.listItem.append(this.listItemContent);

			lastSelected = this.sectionID;
		}


		unselect () {
			this.listItem.classList.remove('selected');
			this.contentSection.classList.remove('selected');

			this.listItemContent.remove();
			this.anchor.append(this.listItemContent);
			this.listItem.append(this.anchor);
		}
	}


	createXHTMLElement('title', {parent: document.query('head'), child: manifest.window.title});

	const refreshButton = getElementById('refresh');
	refreshButton.append(createMenuIcon(refreshButton.title, paths.refresh));
	refreshButton.addEventListener('click', function () {
		const matched = lastSelected.match(/^bookshelf_(\w+)$/);
		if (Array.isArray(matched) && 1 < matched.length) {
			delete networkCache[matched[1]];
		} else {
			networkCache = {};
		}
		location.reload();
	});

	createForm({
		parent: createSection({
			parent: (new TabbedPanel('settings', '全体設定')).contentSection
		}, '初期化'),
		attributes: {method: 'post'},
		child: createParagraph({child: createButton({child: '設定と全ユーザー情報を消去'})})
	}).addEventListener('submit', function (event) {
		if (confirm('設定と全ユーザーの情報を消します。本当によろしいですか?')) {
			localStorage.clear();
			isReset = true;
			return;
		}
		event.preventDefault();
	});


	Object.keys(passwords).forEach(function (username) {
		new TabbedPanel(username);
	});


	const addNewUser = createForm({
		parent: (new TabbedPanel('newUser', '新しいユーザーを登録')).contentSection,
		attributes: {autocomplete: 'off', method: 'post'}
	});
	map2definitionList(new Map([
		[
			createLabel({child: 'ユーザー名', attributes: {htmlFor: 'newUsername'}}),
			createInput({attributes: {id: 'newUsername', maxlength: 128}})
		],
		[
			createLabel({child: 'パスワード', attributes: {htmlFor: 'password4newUser'}}),
			createInput({attributes: {id: 'password4newUser', type: 'password', maxlength: 128}})
		]
	]), {parent: addNewUser});
	const output4newUser = createOutput({parent: createParagraph({parent: addNewUser, children: [
		createButton({child: '登録する'}),
		' ',
		createButton({child: '取り消す', attributes: {type: 'reset'}}),
		' '
	]})});
	addNewUser.addEventListener('submit', function (event) {
		const newUsername = getElementById('newUsername').value;
		const password4newUser = getElementById('password4newUser').value;
		output4newUser.value = '';
		if (!newUsername.length || !password4newUser.length) {
			output4newUser.value = 'ユーザー名またはパスワードが空欄です';
			event.preventDefault();
			return;
		}
		if (newUsername in passwords) {
			if (!confirm(`"${newUsername}" というユーザー${newUsername === nicknames[newUsername] ? '' : ` (表示名: "${nicknames[newUsername]}") `}は既に存在しています。このまま進んだ場合、パスワードをいま入力した値に置き換えることになります。`)) {
				event.preventDefault();
				return;
			}
		} else {
			nicknames[newUsername] = newUsername;
			fileformats[newUsername] = formatTypes[0];
			try {
				filesystem.statSync(path.join(path2bookshelves, newUsername));
			} catch (e) {
				filesystem.mkdirSync(path.join(path2bookshelves, newUsername));
			}
		}
		passwords[newUsername] = password4newUser;
		lastSelected = 'bookshelf_' + newUsername;
	});

	map2definitionList({
		'名称': createXHTMLElement('em', {child: manifest.window.title}),
		'バージョン番号': manifest.version,
		'製作': createAnchor({child: manifest.author.name}, manifest.author.url)
	}, {parent: (new TabbedPanel('about', 'このソフトウェアについて')).contentSection});

	Object.keys(tabs).forEach(function (sectionID) {
		const panel = tabs[sectionID];
		if (sectionID === lastSelected) {
			panel.select();
		} else {
			panel.unselect();
		}
	});
})();
